[galaxy]
server_list=galaxy,automation_hub

[galaxy_server.galaxy]
url={{ fallaxy_galaxy_server }}
token={{ fallaxy_token }}

[galaxy_server.automation_hub]
url={{ fallaxy_ah_server }}
token={{ fallaxy_token }}
