- name: "{{ role_name }} playbook to deprovision the application"
  hosts: localhost
  gather_facts: false
  connection: local
  vars:
    apb_action: deprovision
  roles:
    - role: {{ role_name }}
